/* vector.c
 *
 * by Niels Mller <nisse@lysator.liu.se> 1995
 *
 * Released under the GNU general public licence.
 *
 * Basic operations for vector graphics:
 *
 * Matrix-matrix and matrix-vector multiplication, and generation of
 * a suitable transformation matrix.
 *
 * Coordinates are 3-D, but the transformation matrices work with
 * 4-D homogenous coordinates, to make translation and central projection
 * linear.
 *
 */

#include <math.h>
#include <vectors.h>

/* Compute C = A * B . A, B and C may be the same matrix. */
void MatrixMult(Matrix A, Matrix B, Matrix C)
{
  Matrix temp; /* In case C is the same matrix as A or B, I have to use
		* a temporary matrix for computing, copying the result into
		* C when finished. */
  int i,j,k;

  for (i=0; i<4; i++)
    for (j=0; j<4; j++)
      {
	temp[i][j] = 0;
	for (k=0; k<4; k++)
	  temp[i][j] += A[i][k] * B[k][j];
      }
  memcpy(C, temp, sizeof(Matrix));

#ifdef DBUG
  printf("Matrix product:\n");
  for (i=0; i<4; i++)
    {
      for (j=0; j<4; j++)
	printf("  %lf", C[i][j]);
      printf("\n");
    }
#endif DBUG
}

/* Transforms a list of points */
void TransformVectors(Matrix T, unsigned n, Vector src[], Vector dst[])
{
  HVector temp;
  int i,j;

  for (i=0; i<n; i++)		/* Loop over all vectors */
    {
      for (j=0; j<4; j++)	/* Loop over all the 4 homogenous coordinates */
	/* Last term in sum is implied 4:th homogenous coordinate. */ 
	temp[j] = T[j][0] * src[i][0] + T[j][1] * src[i][1] + T[j][2] * src[i][2] + T[j][3];

      /* Transform back into non-homogenous coordinates. */
      dst[i][0] = temp[0] / temp[3];
      dst[i][1] = temp[1] / temp[3];
      dst[i][2] = temp[2] / temp[3];
    }
}

/* Create transformation matrix from a Viewpoint structure */
void MakeTransform(struct Viewpoint *view, Matrix T)
{
  Matrix R,R1,R2,T1,P; /* Rotation, translation and projection matrices, used below. */
  COORD_T c,s,u,v,ux,uy,w;
  
  /* Set T = translation, to translate the reference point into the origo.
   *
   * The translation is represented as
   *   /        \     / \     /      \
   *  | 1 0 0 dx |   | x |   | x + dx |
   *  | 0 1 0 dy |   | y |   | y + dy |
   *  | 0 0 1 dz | * | z | = | z + dz |
   *  | 0 0 0  1 |   | 1 |   |    1   |
   *   \        /     \ /     \      /
   */

  T[0][0] = 1; T[0][1] = 0; T[0][2] = 0; T[0][3] = -view->ref[0];
  T[1][0] = 0; T[1][1] = 1; T[1][2] = 0; T[1][3] = -view->ref[1];
  T[2][0] = 0; T[2][1] = 0; T[2][2] = 1; T[2][3] = -view->ref[2];
  T[3][0] = 0; T[3][1] = 0; T[3][2] = 0; T[3][3] = 1;   

  /* Length of the direction vector. */
  u = sqrt( SQR(view->direction[0]) + SQR(view->direction[1]) + SQR(view->direction[2]) );
  /* Length of parallell projection into the y-z plane. */
  v = sqrt( SQR(view->direction[1]) + SQR(view->direction[2]) );
  
  /* Rotate about x-axis, to move the view direction into the x-z plane.
   * 
   * The rotation is represented as
   *   /          \ 
   *  | 1  0  0  0 |  
   *  | 0  c  s  0 | 
   *  | 0 -s  c  0 |
   *  | 0  0  0  1 |
   *   \          /
   * where s = sin(phi), c = cos(phi), and phi is the clockwise rotation angle,
   * as seen looking towards the x-axis.
   */
  
  /* v might be very small if we are looking in the x direction,
   * and we don't want to divide by zero, do we? */
  if ( v > EPSILON)
    { 
      s = view->direction[1] / v; c = - view->direction[2] / v;
    }
  else
    {
       s = 0; c = 1;
    }
  
  /* Create the first rotation matrix: */
  R[0][0] =  1; R[0][1] =  0; R[0][2] =  0; R[0][3] =  0;
  R[1][0] =  0; R[1][1] =  c; R[1][2] =  s; R[1][3] =  0;
  R[2][0] =  0; R[2][1] = -s; R[2][2] =  c; R[2][3] =  0;
  R[3][0] =  0; R[3][1] =  0; R[3][2] =  0; R[3][3] =  1;   
  
  /* Rotate about the y-axis, to align the view direction with the negative z-axis.
   *
   * The rotation is represented as
   *   /          \ 
   *  | c  0  s  0 |
   *  | 0  1  0  0 |
   *  |-s  0  c  0 |
   *  | 0  0  0  1 |
   *   \          /
   * where s = sin(psi), c = cos(psi), and psi is the counter-clockwise rotation angle,
   * as seen lookin towards the y-axis.
   */

  s = view->direction[0] / u; c = v / u;
  
  /* Create the rotation matrix: */
  R1[0][0] =  c; R1[0][1] = 0; R1[0][2] =  s; R1[0][3] = 0;
  R1[1][0] =  0; R1[1][1] = 1; R1[1][2] =  0; R1[1][3] = 0;
  R1[2][0] = -s; R1[2][1] = 0; R1[2][2] =  c; R1[2][3] = 0;
  R1[3][0] =  0; R1[3][1] = 0; R1[3][2] =  0; R1[3][3] = 1;

  /* Multiply: R = R1 * R */
  MatrixMult(R1,R,R);

  /* Now determine the x and y components of the rotated up direction. */
  ux = R[0][0] * view->upwards[0] + R[0][1] * view->upwards[1] + R[0][2] * view->upwards[2];
  uy = R[1][0] * view->upwards[0] + R[1][1] * view->upwards[1] + R[1][2] * view->upwards[2];

  w = sqrt( SQR(ux) + SQR(uy) ); /* Length of vector */
  
  /* Rotate about z axis, to align the upward direction with the y axis.
   *
   * The rotation is represented as
   *   /          \ 
   *  | c -s  0  0 |
   *  | s  c  0  0 |
   *  | 0  0  1  0 |
   *  | 0  0  0  1 |
   *   \          /
   * where s = sin(chi), c = cos(chii), and chi is the counter-clockwise rotation angle,
   * as seen lookin towards the z-axis.
   */

  s = ux / w; c = uy / w;

  /* Create rotation matrix: */
  R1[0][0] =  c; R1[0][1] = -s; R1[0][2] =  0; R1[0][3] = 0;
  R1[1][0] =  s; R1[1][1] =  c; R1[1][2] =  0; R1[1][3] = 0;
  R1[2][0] =  0; R1[2][1] =  0; R1[2][2] =  1; R1[2][3] = 0;
  R1[3][0] =  0; R1[3][1] =  0; R1[3][2] =  0; R1[3][3] = 1;

  /* Multiply again: R = R1 * R */
  MatrixMult(R1,R,R);
  
  /* Now compose rotation and translation: T = R * T */
  MatrixMult(R,T,T);

  /* Translate again, so that the projection plane coincides with the x-y plane.
   * This means: subtract view->distance to all the z-coordinates.
   *
   * The translation matrix is:
   *  /               \
   * | 1 0 0 0         |
   * | 0 1 0 0         |
   * | 0 0 1 -distance |
   * | 0 0 0 1         |
   *  \               /
   */

  T1[0][0] = 1; T1[0][1] = 0; T1[0][2] = 0; T1[0][3] = 0;
  T1[1][0] = 0; T1[1][1] = 1; T1[1][2] = 0; T1[1][3] = 0;
  T1[2][0] = 0; T1[2][1] = 0; T1[2][2] = 1; T1[2][3] = -view->distance;
  T1[3][0] = 0; T1[3][1] = 0; T1[3][2] = 0; T1[3][3] = 1;   
  
  /* Apply translation: T = T1 * T */
  MatrixMult(T1,T,T);

  /* Central projection is done with the following matrix:
   *
   *  /             \     / \ 
   * | 1  0 -cx/cz 0 |   | x |
   * | 0  1 -cy/cz 0 |   | y |
   * | 0  0    1   0 | * | z |  , where  cz > 0, and typically cx = cy = 0
   * | 0  0  -1/cz 1 |   | 1 |
   *  \             /     \ / 
   *
   * gives, after transformation into non homogenous coordinates,
   *
   *             cz * (x-cx)
   * x' = cx +  ------------- , similar for y', and
   *               cz - z
   *
   *            cz * z
   * z' =     ---------- 
   *            cz - z
   */

  P[0][0] = 1; P[0][1] = 0; P[0][2] = - view->centre[0]/view->centre[2]; P[0][3] = 0; 
  P[1][0] = 0; P[1][1] = 1; P[1][2] = - view->centre[1]/view->centre[2]; P[1][3] = 0; 
  P[2][0] = 0; P[2][1] = 0; P[2][2] =  1;                                P[2][3] = 0; 
  P[3][0] = 0; P[3][1] = 0; P[3][2] = - 1/view->centre[2];               P[3][3] = 1; 

  /* Compose the projection, and the rest of the transformation. */
  MatrixMult(P,T,T);
}
